import os
import pickle
import time
import matplotlib
import sys
import torch
import numpy as np
from torch.hub import download_url_to_file
import matplotlib.pyplot as plt
import torch.utils.data
import torch.nn.functional as F
import json

from torch.nn.utils.rnn import pack_padded_sequence, PackedSequence, pad_packed_sequence

plt.rcParams["figure.figsize"] = (10, 16) # size of window
plt.style.use('dark_background')

LEARNING_RATE = 1e-3
BATCH_SIZE = 32

EMBEDDING_SIZE = 32
RNN_HIDDEN_SIZE = 128
RNN_LAYERS = 2
RNN_IS_BIDIRECTIONAL = False

TRAIN_TEST_SPLIT = 0.8

MAX_LEN = 200 # limit max number of samples otherwise too slow training (on GPU use all samples / for final training)
DEVICE = 'cpu'

if torch.cuda.is_available():
    DEVICE = 'cuda'
    # comment out this next line if you have nvidia GPU and you want to debug
    MAX_LEN = None
    pass


class Dataset(torch.utils.data.Dataset):
    def __init__(self):
        super().__init__()
        path_dataset = '../data/quotes.pkl'
        if not os.path.exists(path_dataset):
            os.makedirs('../data', exist_ok=True)
            download_url_to_file(
                'http://share.yellowrobot.xyz/1645110979-deep-learning-intro-2022-q1/quotes.pkl',
                path_dataset,
                progress=True
            )

        with open(path_dataset, 'rb') as fp:
            (
                self.final_quotes_sentences, self.final_authors, self.final_categories,
                self.vocabulary_keys, self.vocabulary_counts, self.authors_keys, self.categories_keys
            ) = pickle.load(fp)
        self.max_sentence_length = np.max([len(it) for it in self.final_quotes_sentences])

    def __len__(self):
        if MAX_LEN:
            return MAX_LEN
        return len(self.final_quotes_sentences)

    def __getitem__(self, idx):
        x_raw = np.array(self.final_quotes_sentences[idx])
        y = np.roll(x_raw, -1)
        y = y[:-1]
        x = x_raw[:-1]
        x_length = len(x)
        x_padded = np.pad(x, (0, self.max_sentence_length-x_length))
        y_padded = np.pad(y, (0, self.max_sentence_length-x_length))
        return x_padded, y_padded, x_length

dataset_full = Dataset()

train_test_split = int(len(dataset_full) * TRAIN_TEST_SPLIT)
dataset_train, dataset_test = torch.utils.data.random_split(
    dataset_full,
    [train_test_split, len(dataset_full) - train_test_split],
    generator=torch.Generator().manual_seed(0)
)

dataloader_train = torch.utils.data.DataLoader(
    dataset=dataset_train,
    batch_size=BATCH_SIZE,
    drop_last=(len(dataset_train) % BATCH_SIZE == 1)
)

dataloader_test = torch.utils.data.DataLoader(
    dataset=dataset_test,
    batch_size=BATCH_SIZE,
    drop_last=(len(dataset_test) % BATCH_SIZE == 1)
)

class Model(torch.nn.Module):
    def __init__(self):
        super().__init__()

        self.emb = torch.nn.Embedding(
            num_embeddings=len(dataset_full.vocabulary_keys),
            embedding_dim=EMBEDDING_SIZE
        )

        #TODO different embedding and hidden size
        self.rnn = torch.nn.GRU(
            input_size=EMBEDDING_SIZE,
            batch_first=True,
            hidden_size=EMBEDDING_SIZE,
            num_layers=RNN_LAYERS,
            bidirectional=RNN_IS_BIDIRECTIONAL
        )

    def forward(self, x: PackedSequence, hidden=None):

        embs = self.emb.forward(x.data)
        embs_seq = PackedSequence(
            data=embs,
            batch_sizes=x.batch_sizes,
            sorted_indices=x.sorted_indices
        )
        output, hidden = self.rnn.forward(embs_seq, hidden)

        y_prim_logits = output.data.unsqueeze(dim=-2) @ self.emb.weight.t()
        y_prim_logits = y_prim_logits.squeeze()
        y_prim = torch.softmax(y_prim_logits, dim=-1)
        y_prim_packed = PackedSequence(
            data=y_prim,
            batch_sizes=x.batch_sizes,
            sorted_indices=x.sorted_indices
        )

        return y_prim_packed, hidden


model = Model()
model = model.to(DEVICE)

optimizer = torch.optim.Adam(
    model.parameters(),
    lr=LEARNING_RATE
)

loss_weights = torch.FloatTensor(1.0 - dataset_full.vocabulary_counts / np.sum(dataset_full.vocabulary_counts))
loss_weights = loss_weights.to(DEVICE)

loss_plot_train = []
loss_plot_test = []
acc_plot_train = []
acc_plot_test = []

for epoch in range(1, 1000):

    for dataloader in [dataloader_train, dataloader_test]:
        losses = []
        accs = []
        for x_padded, y_padded, x_length in dataloader:

            x_padded = x_padded.to(DEVICE)
            y_padded = y_padded.to(DEVICE)
            x_length = x_length.to(DEVICE)

            x_packed = pack_padded_sequence(x_padded, x_length, batch_first=True, enforce_sorted=False)
            y_packed = pack_padded_sequence(y_padded, x_length, batch_first=True, enforce_sorted=False)

            y_prim_packed, _ = model.forward(x_packed)

            idxes_batch = range(len(y_packed.data))
            idxes_y = y_packed.data
            loss = -torch.mean(loss_weights[idxes_y] * torch.log(y_prim_packed.data[idxes_batch, idxes_y] + 1e-8))
            losses.append(loss.cpu().item())

            idxes_y_prim = y_prim_packed.data.argmax(dim=-1)
            acc = torch.mean((idxes_y == idxes_y_prim) * 1.0)
            accs.append(acc.cpu().item())

            if dataloader == dataloader_train:
                loss.backward()
                optimizer.step()
                optimizer.zero_grad()

        if dataloader == dataloader_train:
            loss_plot_train.append(np.mean(losses))
            acc_plot_train.append(np.mean(accs))
        else:
            loss_plot_test.append(np.mean(losses))
            acc_plot_test.append(np.mean(accs))

    print(
        f'\n\nepoch: {epoch} '
        f'loss_train: {loss_plot_train[-1]} '
        f'loss_test: {loss_plot_test[-1]} '
        f'acc_plot_train: {acc_plot_train[-1]} '
        f'acc_plot_test: {acc_plot_test[-1]} '
    )

    if epoch % 10 == 0 or True:
        fig, (ax1, ay1) = plt.subplots(2,1)
        ax1.plot(loss_plot_train, 'r-', label='loss train')
        ax1.legend()
        ax1.set_xlabel("Epoch")
        ax2 = ax1.twinx()
        ax2.plot(loss_plot_test, 'c-', label='loss test')
        ax2.legend(loc='upper left')


        ay1.plot(acc_plot_train, 'r-', label='loss acc')
        ay1.legend()
        ay1.set_xlabel("Epoch")
        ay2 = ay1.twinx()
        ay2.plot(acc_plot_test, 'c-', label='loss acc')
        ay2.legend(loc='upper left')

        plt.show()


        x_roll = x_padded[:, :1]
        hidden = torch.zeros((RNN_LAYERS, len(x_roll), EMBEDDING_SIZE))

        for t in range(int(dataset_full.max_sentence_length)):
            lengths = torch.LongTensor([x_roll.size(1)] * len(x_roll))
            x_packed = pack_padded_sequence(x_roll, lengths, batch_first=True)

            y_prim, hidden = model.forward(x_packed, hidden)

            y_prim, _ = pad_packed_sequence(y_prim, batch_first=True)
            y_prim_idx = y_prim[:, -1:].argmax(-1)
            x_roll = torch.cat((x_roll, y_prim_idx), dim=-1)

        np_x_roll = x_roll.cpu().data.numpy().tolist()
        for each_x in np_x_roll:
            sent = [dataset_full.vocabulary_keys[it] for it in each_x]
            print(' '.join(sent))